open Board

type 'a points = (int,(int,'a) Hashtbl.t) Hashtbl.t

let create_points () = Hashtbl.create 17

let add_point points data x y =
  try
    let col = Hashtbl.find points x in
    Hashtbl.replace col y data
  with
    Not_found -> 
      let col = Hashtbl.create 17 in
      Hashtbl.add col y data;
      Hashtbl.add points x col
    
let find_points points x y =
  let col = Hashtbl.find points x in
  Hashtbl.find col y 

let mem_points points x y =
  try
    let _ = find_points points x y in
    true
  with
    Not_found -> false

let print_points fmt points =
  Hashtbl.iter
    (fun x col ->
      Hashtbl.iter 
	(fun y data ->
	  Format.printf " (%d,%d)" x y)
	col)
    points

let accessible x y =
  match board.(x).(y) with
  | Wall | Water -> false
  | Plain | Home | Robot -> true
	    
let visited = Array.create_matrix (width+1) (height+1) false
let clear_visited () =
  for c = 0 to width do Array.fill visited.(c) 0 (height+1) false done

type path =
  | Full of direction * direction * direction
  | Two of direction * direction
  | One of direction
  | Empty

let add_to_path d = function
  | Full _ as p -> p
  | One d' -> Two (d', d)
  | Two(d1,d2) -> Full(d1,d2,d)
  | Empty -> assert false

exception Dir_found of path * int * int * int;;

let board_nearest x y points = 
  clear_visited ();
  let to_visit = Queue.create () in
  visited.(x).(y) <- true;
  try
    List.iter
      (fun (d,x1,y1) ->
	if accessible x1 y1 then
	  if mem_points points x1 y1 then
	    raise (Dir_found (d,1,x1,y1))
	  else 
	    begin
	      visited.(x1).(y1) <- true;
	      Queue.add (d,1,x1,y1) to_visit
	    end)
      [(One North,x,y+1);(One East,x+1,y);(One South,x,y-1);(One West,x-1,y)];
    while true do 
      (*
      Format.printf "Queue is :";
      Queue.iter 
	(fun (d,x,y) ->
	Format.printf " (%c,%d,%d)" (char_of_direction d) x y) 
	to_visit;
      Format.printf "@.";
      *)
      let (p,dist,x,y) = Queue.take to_visit in
      List.iter
	(fun (d,x1,y1) ->
	  if accessible x1 y1 then
	    if mem_points points x1 y1 then
	      raise (Dir_found (add_to_path d p,dist,x1,y1))
	    else 
	      if not (visited.(x1).(y1)) then 
		begin
		  visited.(x1).(y1) <- true;
		  Queue.add (add_to_path d p,succ dist,x1,y1) to_visit
		end)
      [(North,x,y+1);(East,x+1,y);(South,x,y-1);(West,x-1,y)];
    done;
    assert false
  with
  | Dir_found (p,dist,x,y) -> 
      (*
	(match p with One _ -> prerr_endline "One"
	| Full _ -> prerr_endline "Full"
	| Empty -> assert false);
      *)
      (find_points points x y,p,dist)
  | Queue.Empty -> raise Not_found
;;
	
