
(* ICFP programming contest 2002 *)

open Board;;
open Strategy;;
open Format

(* action returns a value of type [bid * action] *)

let nact = ref 0

let action () =
  (*i dummy i*)
  let t = (Unix.times ()).Unix.tms_utime in
  let a = neanderthal () in
  begin match a with 
    | _, Move North -> next_position := (self.x, self.y + 1)
    | _, Move South -> next_position := (self.x, self.y - 1)
    | _, Move East  -> next_position := (self.x + 1, self.y)
    | _, Move West  -> next_position := (self.x - 1, self.y)
    | _ -> ()
  end;
  let dt = (Unix.times ()).Unix.tms_utime -. t in
  (* Format.printf "action %d sent in %f.@." !nact dt; *)
  a
    
(* sending the action to the server *)

let rec print_packages fmt = function
  | [] -> ()
  | [p] -> fprintf fmt "%d" p
  | p :: pl -> fprintf fmt "%d %a" p print_packages pl

let print_action fmt = function
  | Move North -> fprintf fmt "Move N"
  | Move South -> fprintf fmt "Move S"
  | Move East -> fprintf fmt "Move E"
  | Move West -> fprintf fmt "Move W"
  | Pick pl -> fprintf fmt "Pick %a" print_packages pl
  | Drop pl -> fprintf fmt "Drop %a" print_packages pl

let action_buffer = Buffer.create 256
let bfmt = formatter_of_buffer action_buffer

let send_action (bid, a) =
  Buffer.clear action_buffer;
  fprintf bfmt "%d %a@?" bid print_action a;
  Connect.send_line (Buffer.contents action_buffer)

(* interpreting packages' list *)

let interp_packages s =
  let rec interp = function
    | p :: dx :: dy :: w :: r ->
	(* Claude : replaced Hashtbl.add by Hashtbl.replace *)
	Hashtbl.replace packages p { dest_x = dx; dest_y = dy; weight = w };
	available.(self.x).(self.y) <- Ptset.add p available.(self.x).(self.y);
	Hashtbl.replace available_packages p (self.x,self.y);
	interp r
    | [] -> 
	()
    | _ -> 
	assert false
  in
  available.(self.x).(self.y) <- Ptset.empty;
  let l =
    Hashtbl.fold
      (fun p (x,y) acc ->
	if x = self.x && y = self.y then p::acc else acc)
      available_packages
      []
  in
  List.iter (fun p -> Hashtbl.remove available_packages p) l;
  interp (split_ints s)

(* game loop *)

let rec game_loop () =
  (* 1. first wait for the packages list *)
  let s = Connect.read_line () in
  (*i 
  prerr_endline ("** server packages : [" ^ s ^ "]");
  i*)
  if s <> "" && String.get s 0 = 'R' then
    prerr_endline ("** server tells me : [" ^ s ^ "]")
  else interp_packages s;
  (* 2. send action *)
  let (bid,dec) as a = action () in
  (*i
     printf "** action : %a@." print_action (snd a);
   i*)
  (* Claude : decremente money *)
  if (abs bid) > !money then
    begin
      send_action (!money,dec);
      money := 0
    end
  else
    begin
      send_action a;
      money := !money - (abs bid);
    end;
  old2_x := !old1_x;
  old2_y := !old1_y;
  old1_x := self.x;
  old1_y := self.y;
  old2_dec := !old1_dec;
  old1_dec := dec;
  (* 3. get server reply *)
  let s = Connect.read_line () in
  (*i 
  prerr_endline ("** server answer : [" ^ s ^ "]");
  i*)
  interp_reply s;
  if (self.x, self.y) <> !next_position then nothing_changed := false;
  game_loop ()

(* debugging *)
let dump_state () =
  printf "cx = %d  cy = %d  capacity = %d  money = %d\n" 
    self.x self.y !capacity !money;
  let print_package i info =
    printf "package %d: destination = %d, %d  weight = %d\n" 
      i info.dest_x info.dest_y info.weight
  in
  Hashtbl.iter print_package packages; printf "\n";
  printf "END\n"

let _ = Sys.catch_break true


let _ = 
  try game_loop () 
  with 
  | End_of_file | Unix.Unix_error _ -> 
      (* dump_state(); *)
      Format.printf "Connection closed by server@.";
      let s = ref 0 in
      Hashtbl.iter 
	(fun pid pinfo -> s := !s + pinfo.weight)
	packages;
      Format.printf "Maximum score was (as far as I know) %d@." !s;      
  | Exit -> ()
(*  | e -> 
      dump_state (); raise e
*)
