
(* strategy *)

open Board;;
open Algo;;

type action = 
  | Move of direction 
  | Pick of int list
  | Drop of int list

type bid = int

(* dummy : le robot qui ne fait jamais rien *)

let dummy () = (1, Drop [])

(* cromagnon : delivre le vite possible *)

exception Decide of int * action
    
let bases_connues = create_points ()
    

let estim_score = ref 0

let rec add_in_pid_list w p l =
  match l with
  | [] -> [(w,p)]
  | ((w1,p1) as a)::r ->
      if w <= w1 then (w,p)::l (* petits paquets en premier *)
      else a::(add_in_pid_list w p r)

let rec add_in_ord_list pid weight dest_x dest_y acc =
  match acc with
  | [] -> [(weight,[(weight,pid)],dest_x,dest_y)]
  | ((w,l,x,y) as a)::r ->
      if x=dest_x && y=dest_y then
	(w+weight,add_in_pid_list weight pid l,x,y)::r
      else
	a::(add_in_ord_list pid weight dest_x dest_y r)

let cromagnon_select_packages avail capac =
 
  (* ordlist : liste des (poids,pids,x,y) des paquets disponibles,
     regroupes par destination identiques, et par poids total decroissant *)

  let ordlist =
    Ptset.fold
      (fun pid acc ->
	try
	  let pinfo = Hashtbl.find packages pid in
	  add_in_ord_list pid pinfo.weight pinfo.dest_x pinfo.dest_y acc
	with Not_found -> raise Exit (* assert false *))
      avail
      []
  in
  let by_weight = 
    List.sort (fun (w1,_,_,_) (w2,_,_,_) -> compare w2 w1) ordlist
  in
  let (take,_) =
    List.fold_left
      (fun acc (_,pidl,_,_) ->
	List.fold_left
	  (fun ((l,c) as acc) (w,p) -> 
	    if w <= c then (p::l,c-w) else acc)
	  acc
	  pidl)
      ([],capac)
      by_weight
  in take



  


let current_path = ref Empty
let delivery_dist = ref 0
let high_bid = ref 10

let get_dir = function
  | Empty -> 
      raise Exit (* assert false *)
  | One d -> 
      current_path := Empty; d
  | Two (d1, d2) -> 
      current_path := One d2; d1
  | Full (d1, d2, d3) -> 
      current_path := Two(d2,d3); d1

let cromagnon ()=
  try
    (* warning before death ! *)
    if !money <= 1 then
      Format.printf "%d zorkmid left, I shall die soon...@." !money;
    (* mise a jour bases connues si necessaire *)
    if board.(self.x).(self.y) = Home && 
      not(mem_points bases_connues self.x self.y) then
      add_point bases_connues () self.x self.y;

    (* y a-t-il des paquets a poser ici ? *)
    (**************************************)
    let a_poser =
      Ptset.fold
	(fun pid acc ->
	  let pinfo = Hashtbl.find packages pid in
	  if self.x = pinfo.dest_x &&  self.y = pinfo.dest_y then
	    pid::acc
	  else 
	    acc)
	self.load
	[]
    in
    if a_poser <> [] then
      begin
      (* oui, on droppe *)
	List.iter 
	  (fun pid ->
	    let pinfo = Hashtbl.find packages pid in
	    estim_score := !estim_score + pinfo.weight
	    (*
	      Format.printf 
	      "Dropping package %d, new score (if nobody pushes me) = %d@." pid !estim_score*))
	      a_poser;
	    
	raise (Decide (1, Drop a_poser))
      end;

    (* y a-t-il des paquets a ramasser ici ? *)
    (*****************************************)
    let avail = available.(self.x).(self.y) in
    let a_prendre = 
      cromagnon_select_packages avail !capacity 
    in
    (* on ramasse *)
    if a_prendre <> [] then
      begin
	(*i
	List.iter 
	  (fun pid -> Format.printf "Picking package %d@." pid)
	  a_prendre;
	   i*)
	raise (Decide(1,Pick a_prendre))
      end;

    (* si rien n'a change et on a un chemin memorise on le prend *)
    if !nothing_changed then begin
      match !current_path with
	| Empty -> ()
	| p -> (* prerr_endline "on continue"; *)
	    raise (Decide (1, Move (get_dir p)))
    end;

    (* on cherche a livrer le paquet qui a la destination la plus proche *)
    (*********************************************************************)
    if not(Ptset.is_empty self.load) then
      begin
	let dests = create_points () in
	Ptset.iter
	  (fun pid ->
	    try
	      let pinfo = Hashtbl.find packages pid in
	      add_point dests pid pinfo.dest_x pinfo.dest_y
	  with Not_found -> raise Exit (* assert false ??? *))
	  self.load;
	try
	  let (pid,p,dist) = board_nearest self.x self.y dests in
	  if !high_bid > !money - 2*dist
	  then high_bid := max 2 (!money - 2*dist);
	  raise (Decide (1, Move (get_dir p)))
	  (*i
	     Format.printf "Je vais livrer le paquet %d@." pid;
	     i*)
	with
	  Not_found -> 
	      (* aucune destination accessible ! *)
	    raise 
	      (Decide
		 (1,Drop 
		    (Ptset.fold (fun x acc -> x::acc) self.load [])))
      end;


    (* on n'a rien a livrer, on cherche a aller vers un endroit ou
       prendre un paquet *)

    let dests = create_points () in
    List.iter
      (fun (x,y) -> 
	if not(mem_points bases_connues x y) then
	  add_point dests (-1) x y)
      bases;
    Hashtbl.iter
      (fun pid (px,py) ->
	add_point dests pid px py)
      available_packages;
    (*i
    Format.printf 
      "Je vais chercher un paquet vers les cases suivantes : %a@."
      print_points dests;
       i*)
    begin
      try
	let (pid,p,dist) = board_nearest self.x self.y dests in
	(*i
	if pid >= 0 then 
	  Format.printf "Je vais chercher le paquet %d@." pid
	else
	  Format.printf "Je vais vers une base@.";
	   i*)
	raise (Decide(1,Move (get_dir p)))
      with
	Not_found -> 
	  (* aucune package accessible ! *)
	  ()
    end;

    (* en dernier ressort !*)

    (1,Drop [])
     
with
  | Decide(bid,act) -> (bid,act)
  | _ -> (* si probleme *) (1,Drop [])
;;

(* end of cromagnon *)


(* neanderthal *)

(*

  recherche d'une situation dangereuse

*)

let board_content x y =
  let r = robots.(x).(y) in
  if r = 0 then board.(x).(y) else Robot


type rel_move = Left | Right | Forw

let check_move_evasive_action_aux d x y =
  let forw,left,right,forw_left,forw_right =
  match d with 
    | North -> 
	board_content x (succ y),
	board_content (pred x) y,
	board_content (succ x) y,
	board_content (pred x) (succ y),
	board_content (succ x) (succ y)
    | South -> 
	board_content x (pred y),
	board_content (succ x) y,
	board_content (pred x) y,
	board_content (succ x) (pred y),
	board_content (pred x) (pred y)
    | East -> 
	board_content (succ x) y,
	board_content x (succ y),
	board_content x (pred y),
	board_content (succ x) (succ y),
	board_content (succ x) (pred y)
    | West -> 
	board_content (pred x) y,
	board_content x (pred y),
	board_content x (succ y),
	board_content (pred x) (pred y),
	board_content (pred x) (succ y)
  in
  match left,right,forw_left,forw_right with
    | Water,Robot,Water,Robot 
    | Water,Robot,Robot,Water -> !high_bid,Right
    | Robot,Water,Water,Robot 
    | Robot,Water,Robot,Water -> !high_bid,Left
    | Water,Robot,_,_
    | Robot,Water,_,_ -> !high_bid,Forw
    | _,_,Robot,Water 
    | _,_,Water,Robot -> (- !high_bid),Forw
    | _ -> 1,Forw
  
let compute_move rel d =
  match (rel,d) with
    | (Forw,d) -> d
    | (Left,North) -> West
    | (Right,North) -> East
    | (Left,South) -> East
    | (Right,South) -> West
    | (Left,East) -> North
    | (Right,East) -> South
    | (Left,West) -> South
    | (Right,West) -> North

let check_move_evasive_action d x y =
  let bid,rel = check_move_evasive_action_aux d x y in
  (bid,Move (compute_move rel d))

let check_stay_evasive_action dec x y =
  let north = board_content x (succ y)
  and south = board_content x (pred y)
  and east = board_content (succ x) y
  and west = board_content (pred x) y
  in
  match north,south,east,west with
      (* robot au sud et eau au nord *)
    | Water,Robot,(Water|Wall),(Water|Wall) -> !high_bid,Move South 
    | Water,Robot,(Water|Wall),_ -> check_move_evasive_action West x y
    | Water,Robot,_,(Water|Wall) -> check_move_evasive_action East x y
    | Water,Robot,_,_ ->
	let dec_east = check_move_evasive_action East x y
	and dec_west = check_move_evasive_action West x y
	in
	if fst dec_east > 0 then dec_east
	else if fst dec_west > 0 then dec_west
	else !high_bid,Move South
 
      (* robot au nord et eau au sud *)
    | Robot,Water,(Water|Wall),(Water|Wall) -> !high_bid,Move North 
    | Robot,Water,(Water|Wall),_ -> check_move_evasive_action West x y
    | Robot,Water,_,(Water|Wall) -> check_move_evasive_action East x y
    | Robot,Water,_,_ ->
	let dec_east = check_move_evasive_action East x y
	and dec_west = check_move_evasive_action West x y
	in
	if fst dec_east > 0 then dec_east
	else if fst dec_west > 0 then dec_west
	else !high_bid,Move North
 
      (* robot a l'est et eau a l'ouest *)
    | (Water|Wall),(Water|Wall),Robot,Water -> !high_bid,Move East 
    | (Water|Wall),_,Robot,Water -> check_move_evasive_action South x y
    | _,(Water|Wall),Robot,Water -> check_move_evasive_action North x y
    | _,_,Robot,Water ->
	let dec_north = check_move_evasive_action North x y
	and dec_south = check_move_evasive_action West x y
	in
	if fst dec_north > 0 then dec_north
	else if fst dec_south > 0 then dec_south
	else !high_bid,Move East

      (* robot a l'ouest et eau a l'est *)
    | (Water|Wall),(Water|Wall),Water,Robot -> !high_bid,Move West 
    | (Water|Wall),_,Water,Robot -> check_move_evasive_action North x y
    | _,(Water|Wall),Water,Robot -> check_move_evasive_action South x y
    | _,_,Water,Robot ->
	let dec_north = check_move_evasive_action North x y
	and dec_south = check_move_evasive_action West x y
	in
	if fst dec_north > 0 then dec_north
	else if fst dec_south > 0 then dec_south
	else !high_bid,Move West


    | _ -> 1,dec


let old2_dec = ref (Drop [])

let old1_dec = ref (Drop [])

let neanderthal () =
  high_bid := !money / 10;
  let (_,dec) = cromagnon () in
(*
  high_bid := 1;
*)
  let final_dec =
    match dec with
      | Move d -> 
	  check_move_evasive_action d self.x self.y
      | Pick _ | Drop _ -> 
	  check_stay_evasive_action dec self.x self.y 
  in
  if !old2_x = self.x && !old2_y = self.y && !old2_dec = snd final_dec &&
     (match snd final_dec with Move _ -> true | _ -> false) &&
     Random.int 3 = 0
  then (fst final_dec,Drop [])
  else final_dec



